# -*- coding: utf-8 -*-

"""
supercell_step
A step for building supercells of periodic systems.
"""

# Bring up the classes so that they appear to be directly in
# the supercell_step package.

from supercell_step.supercell import Supercell  # noqa: F401, E501
from supercell_step.supercell_parameters import SupercellParameters  # noqa: F401, E501
from supercell_step.supercell_step import SupercellStep  # noqa: F401, E501
from supercell_step.tk_supercell import TkSupercell  # noqa: F401, E501

# Handle versioneer
from ._version import get_versions

__author__ = """Paul Saxe"""
__email__ = 'psaxe@molssi.org'
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



# -*- coding: utf-8 -*-

"""The graphical part of a Supercell step"""

import seamm
import seamm_widgets as sw
import tkinter as tk


class TkSupercell(seamm.TkNode):
    """
    The graphical part of a Supercell step in a flowchart.

    Attributes
    ----------
    namespace : str
        The namespace of the current step.
    node : Node
        The corresponding node of the non-graphical flowchart
    dialog : Dialog
        The Pmw dialog object
    sub_tk_flowchart : TkFlowchart
        A graphical Flowchart representing a subflowchart
    self[widget] : dict
        A dictionary of tk widgets built using the information
        contained in Supercell_parameters.py

    See Also
    --------
    Supercell, TkSupercell,
    SupercellParameters,
    """

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50
    ):
        """
        Initialize a graphical node.

        Parameters
        ----------
        tk_flowchart: Tk_Flowchart
            The graphical flowchart that we are in.
        node: Node
            The non-graphical node for this step.
        namespace: str
            The stevedore namespace for finding sub-nodes.
        canvas: Canvas
           The Tk canvas to draw on.
        x: float
            The x position of the nodes center on the canvas.
        y: float
            The y position of the nodes cetner on the canvas.
        w: float
            The nodes graphical width, in pixels.
        h: float
            The nodes graphical height, in pixels.
        """
        self.dialog = None

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h
        )

    def create_dialog(self):
        """
        Create the dialog. A set of widgets will be chosen by default
        based on what is specified in the
        Supercell_parameters module.

        See Also
        --------
        TkSupercell.reset_dialog
        """

        super().create_dialog(title='Edit Supercell step')

        # The information about widgets is held in self['xxxx'], i.e. this
        # class is in part a dictionary of widgets. This makes accessing
        # the widgets easier and allows loops, etc.

        # Shortcut for parameters
        P = self.node.parameters

        # Then create the widgets
        for key in P:
            self[key] = P[key].widget(self['frame'])

        # and lay them out
        self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Layout the widgets in the dialog.

        The widgets are chosen by default from the information in
        Supercell_parameter.

        This function simply lays them out row by row with
        aligned labels. You may wish a more complicated layout that
        is controlled by values of some of the control parameters.

        Parameters
        ----------
        widget

        See Also
        --------
        TkSupercell.create_dialog
        """

        # Remove any widgets previously packed
        frame = self['frame']
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Shortcut for parameters
        P = self.node.parameters

        # keep track of the row in a variable, so that the layout is flexible
        # if e.g. rows are skipped to control such as 'method' here
        row = 0
        widgets = []
        for key in P:
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        # Align the labels
        sw.align_labels(widgets)

    def right_click(self, event):
        """
        Handles the right click event on the node.

        See Also
        --------
        TkSupercell.edit
        """

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def edit(self):
        """Present a dialog for editing the Supercell input

        See Also
        --------
        TkSupercell.right_click
        """

        if self.dialog is None:
            self.create_dialog()

        self.dialog.activate(geometry='centerscreenfirst')

    def handle_dialog(self, result):
        """Handle the closing of the edit dialog

        What to do depends on the button used to close the dialog. If
        the user closes it by clicking the 'x' of the dialog window,
        None is returned, which we take as equivalent to cancel.

        Parameters
        ----------
        result : None or str
            The value of this variable depends on what the button
            the user clicked.

        """

        if result is None or result == 'Cancel':
            self.dialog.deactivate(result)
            return

        if result == 'Help':
            # display help!!!
            return

        if result != "OK":
            self.dialog.deactivate(result)
            raise RuntimeError(
                "Don't recognize dialog result '{}'".format(result)
            )

        self.dialog.deactivate(result)
        # Shortcut for parameters
        P = self.node.parameters

        # Get the values for all the widgets. This may be overkill, but
        # it is easy! You can sort out what it all means later, or
        # be a bit more selective.
        for key in P:
            P[key].set_from_widget()

    def handle_help(self):
        """Shows the help to the user when click on help button.

        """
        print('Help not implemented yet for Supercell!')



# -*- coding: utf-8 -*-

"""Non-graphical part of the Supercell step in a SEAMM flowchart
"""

import logging
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import supercell_step

# In addition to the normal logger, two logger-like printing facilities are
# defined: 'job' and 'printer'. 'job' send output to the main job.out file for
# the job, and should be used very sparingly, typically to echo what this step
# will do in the initial summary of the job.
#
# 'printer' sends output to the file 'step.out' in this steps working
# directory, and is used for all normal output from this step.

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('Supercell')


class Supercell(seamm.Node):
    """
    The non-graphical part of a Supercell step in a flowchart.

    Attributes
    ----------
    options : tuple
        It contains a two item tuple containing the populated namespace and the
        list of remaining argument strings.

    subflowchart : seamm.Flowchart
        A SEAMM Flowchart object that represents a subflowchart, if needed.

    parameters : SupercellParameters
        The control parameters for Supercell.

    See Also
    --------
    TkSupercell,
    Supercell, SupercellParameters
    """

    def __init__(
        self,
        flowchart=None,
        title='Supercell',
        extension=None,
        logger=logger
    ):
        """A step for Supercell in a SEAMM flowchart.

        You may wish to change the title above, which is the string displayed
        in the box representing the step in the flowchart.

        Parameters
        ----------
            flowchart: seamm.Flowchart
                The non-graphical flowchart that contains this step.

            title: str
                The name displayed in the flowchart.
            extension: None
                Not yet implemented
        """
        super().__init__(
            flowchart=flowchart,
            title='Supercell',
            extension=extension,
            logger=logger
        )

        self.parameters = supercell_step.SupercellParameters()

    @property
    def version(self):
        """The semantic version of this module.
        """
        return supercell_step.__version__

    @property
    def git_revision(self):
        """The git version of this module.
        """
        return supercell_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.

        Parameters
        ----------
            P: dict
                An optional dictionary of the current values of the control
                parameters.
        Returns
        -------
            description : str
                A description of the current step.
        """

        if not P:
            P = self.parameters.values_to_dict()

        text = ('Create a {na} x {nb} x {nc} supercell from the current cell')

        return self.header + '\n' + __(text, **P, indent=4 * ' ').__str__()

    def run(self):
        """Create the supercell.

        The strategy used is to expand in the `a` direction, adding the atoms
        and bonds from the original system for each cell add, and keeping the
        new fractional coordinates. Then move to the `b` direction, getting the
        current atoms and bonds for the expanded `a` direction. And finally for
        the `c` direction. At the end, after setting the cell the coordinates
        are updated with the fractional coordinates that have been accumulated.

        Returns
        -------
        next_node : seamm.Node
            The next node object in the flowchart.

        """

        next_node = super().run(printer)
        # Get the values of the parameters, dereferencing any variables
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Print what we are doing
        printer.important(__(self.description_text(P), indent=self.indent))

        # Get the current system
        system_db = self.get_variable('_system_db')
        configuration = system_db.system.configuration

        atoms = configuration.atoms
        bonds = configuration.bonds
        cell = configuration.cell

        na = P['na']
        nb = P['nb']
        nc = P['nc']
        logger.debug(f"making {na} x {nb} x {nc} supercell")

        # Get a copy of the initial atom and bond data
        atom_data = atoms.get_as_dict()
        # index of atoms to use for bonds
        index = {j: i for i, j in enumerate(atom_data['id'])}
        del atom_data['id']
        bond_data = bonds.get_as_dict()
        del bond_data['id']

        # Get the initial fractional coordinates, adjusting to the final cell
        xyz0 = [
            [x / na, y / nb, z / nc]
            for x, y, z in atoms.get_coordinates(fractionals=True)
        ]
        xyzs = list(xyz0)

        # Expand the cell along 'a'
        for ia in range(1, na):
            # Coordinates
            for x, y, z in xyz0:
                xyzs.append([x + ia / na, y, z])
            # Atoms
            ids = atoms.append(**atom_data)
            # Bonds
            bond_data['i'] = [ids[index[i]] for i in bond_data['i']]
            bond_data['j'] = [ids[index[j]] for j in bond_data['j']]
            bonds.append(**bond_data)

        # Get a copy of the current atom and bond data
        atom_data = atoms.get_as_dict()
        # index of atoms to use for bonds
        index = {j: i for i, j in enumerate(atom_data['id'])}
        del atom_data['id']
        bond_data = bonds.get_as_dict()
        del bond_data['id']

        # Keep a copy of the current coordinates
        xyz0 = list(xyzs)

        # Expand the cell along 'b'
        for ib in range(1, nb):
            # Coordinates
            for x, y, z in xyz0:
                xyzs.append([x, y + ib / nb, z])
            # Atoms
            ids = atoms.append(**atom_data)
            # Bonds
            bond_data['i'] = [ids[index[i]] for i in bond_data['i']]
            bond_data['j'] = [ids[index[j]] for j in bond_data['j']]
            bonds.append(**bond_data)

        # Get a copy of the current atom and bond data
        atom_data = atoms.get_as_dict()
        # index of atoms to use for bonds
        index = {j: i for i, j in enumerate(atom_data['id'])}
        del atom_data['id']
        bond_data = bonds.get_as_dict()
        del bond_data['id']

        # Keep a copy of the current coordinates
        xyz0 = list(xyzs)

        # Expand the cell along 'c'
        for ic in range(1, nc):
            # Coordinates
            for x, y, z in xyz0:
                xyzs.append([x, y, z + ic / nc])
            # Atoms
            ids = atoms.append(**atom_data)
            # Bonds
            bond_data['i'] = [ids[index[i]] for i in bond_data['i']]
            bond_data['j'] = [ids[index[j]] for j in bond_data['j']]
            bonds.append(**bond_data)

        # Update the cell
        a, b, c, alpha, beta, gamma = cell.parameters
        logger.debug(f"initial cell: {a}, {b}, {c}, {alpha}, {beta}, {gamma}")
        a *= na
        b *= nb
        c *= nc
        logger.debug(f"final cell: {a}, {b}, {c}, {alpha}, {beta}, {gamma}")
        cell.parameters = (a, b, c, alpha, beta, gamma)

        # Set all the coordinates using the new cell
        atoms.set_coordinates(xyzs, fractionals=True)

        # Print what we did
        n_atoms = atoms.n_atoms
        printer.important(
            __(
                (
                    f'Created a {na} x {nb} x {nc} supercell containing '
                    f'{n_atoms} atoms with cell parameters:'
                ),
                indent=self.indent + 4 * ' ',
            )
        )
        tmp = self.indent + 8 * ' '
        printer.important('')
        printer.important(tmp + f'    a = {a:8.3f} Å')
        printer.important(tmp + f'    b = {b:8.3f}')
        printer.important(tmp + f'    c = {c:8.3f}')
        printer.important(tmp + f'alpha = {alpha:7.2f} degrees')
        printer.important(tmp + f' beta = {beta:7.2f}')
        printer.important(tmp + f'gamma = {gamma:7.2f}')
        printer.important('')

        # Analyze the results
        self.analyze()

        return next_node



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `supercell_step` package."""

import pytest  # noqa: F401
import supercell_step  # noqa: F401
import re


@pytest.fixture
def instance():
    instance = supercell_step.Supercell()
    instance._id = (1,)
    return instance


def test_construction():
    """Simplest test that we can make a SUPERCELL object"""
    instance = supercell_step.Supercell()
    assert (
        str(type(instance)) == "<class 'supercell_step.supercell.Supercell'>"
    )


def test_version():
    """Test that the object returns a version"""
    instance = supercell_step.Supercell()
    result = instance.version
    assert isinstance(result, str) and len(result) > 0


def test_git_revision():
    """Test that the object returns a git revision"""
    instance = supercell_step.Supercell()
    result = instance.git_revision
    assert isinstance(result, str) and len(result) > 0


def test_description_text_default(instance):
    """Test the default description text"""

    assert re.fullmatch(
        (
            r'Step 1: Supercell  [-+.0-9a-z]+\n'
            r'    Create a 2 x 2 x 2 supercell from the current cell'
        ), instance.description_text()
    ) is not None


def test_description_text_expr_expr(instance):
    """Test the default description text"""

    assert re.fullmatch(
        (
            r'Step 1: Supercell  [-+.0-9a-z]+\n'
            r'    Create a 5 x 4 x 3 supercell from the current cell'
        ), instance.description_text({
            'na': 5,
            'nb': 4,
            'nc': 3,
        })
    ) is not None
